/* -*- knf -*- */
/* $AMSEL: main.c,v 1.8 2004/05/23 15:03:38 reyk Exp $ */

/*
 * Copyright (c) 2004 .vantronix | secure systems GmbH
 * Copyright (c) 2004 SSV Embedded Systems GmbH
 *
 * All rights reserved.
 *
 * Written by Reyk Floeter <reyk@vantronix.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  *
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "main.h"
#include "tables.h"
#include "ipvs.h"
#include "parser.h"
#include "version.h"

#define _GNU_SOURCE 1
#include <getopt.h>

int errno;

struct amselwall_config *conf = NULL;
int pdebug = 0;
char *amwall_me = NULL;

static void amwall_usage(char *);
static int amwall_getopt(int, char *[]);

static void
amwall_usage(char *name)
{
	printf(VERSION_NAME " (build %s)\n", VERSION_CHR, VERSION_BUILD);
	printf("Usage: %s [OPTIONS]...\n"
	    "  -h, --help        Show help\n"
	    "  -f, --file file   Set configuration file (%s)\n"
	    "  -F, --flush       Generate flushing rules\n"
	    "  -k, --keep        Keep existing rules, don't flush\n"
	    "  -e, --edit        Try to edit/replace existing rules\n"
	    "  -O, --optimize    Try to optimize calls, rules, ...\n"
	    "  -q, --quiet       Be quiet, parse input file and quit\n"
	    "  -v, --verbose     Be verbose\n",
	    name, AMSELWALL_DEFAULT_CONFIG);
}

static int
amwall_getopt(int argc, char *argv[])
{
	int ch = 0;
	static struct option long_options[] = {
		{ "help", 0, NULL, 'h' },
		{ "quiet", 0, NULL, 'q' },
		{ "verbose", 0, NULL, 'v' },
		{ "edit", 0, NULL, 'e' },
		{ "optimize", 0, NULL, 'O' },
		{ "keep", 0, NULL, 'k' },
		{ "flush", 0, NULL, 'F' },
		{ "file", 1, NULL, 'f' },
		{ NULL, }
	};
	
	if(argv == NULL || argc == 0)
		return(0);
	
	while (1) {
		int option_index = 0;
		
		if((ch = getopt_long(argc, argv, "hq?veOkFf:", long_options,
			&option_index)) == -1)
			break;

		switch(ch) {
		case 'h':
		case '?':
			amwall_usage(argv[0]);
			return(-EINVAL);
			break;

		case 'f':
			if((conf->path = strdup(optarg)) == NULL)
				fatal("out of memory");
			conf->flags |= AMSELWALL_OPT_PATH;
			break;

		case 'v':
			conf->flags |= AMSELWALL_OPT_VERBOSE;
			break;

		case 'O':
			conf->flags |= AMSELWALL_OPT_OPTIMIZE;
			break;

		case 'F':
			conf->flags |= AMSELWALL_OPT_FLUSH;
			break;

		case 'e':
			conf->flags |= AMSELWALL_OPT_EDIT;
			break;

		case 'k':
			conf->flags |= AMSELWALL_OPT_KEEP;
			break;

		case 'q':
			conf->flags |= AMSELWALL_OPT_QUIET;
			break;

		default:
			fprintf(stderr, "Invalid option '%c'\n", ch);
			amwall_usage(argv[0]);
			return(-EINVAL);
		}
	}

	return(optind);
}

static void
amwall_conf_init(void)
{
	conf->path = NULL;
	TAILQ_INIT(&conf->h_head);
}

static void
amwall_conf_free(void)
{
	if(conf) {
		amwall_cleanup();

		if(conf->path)
			free(conf->path);

		free(conf);
	}
}

static int
amwall_lookup_cmp(const void *k, const void *e)
{
	return(strcasecmp(k, ((const struct keywords *)e)->k_name));
}

int
amwall_lookup_keyword(char *s, const struct keywords *table, size_t nmemb)
{
	const struct keywords *p;

	p = bsearch(s, table, nmemb, sizeof(struct keywords), amwall_lookup_cmp);
	
	if(p != NULL)
		return(p->k_val);

	return(-ENOENT);
}

int 
main(int argc, char *argv[])
{
	int ret;
	struct amselwall_table *table;

	amwall_me = argv[0];

	signal(SIGPIPE, SIG_IGN);

	if(atexit(amwall_conf_free) != 0)
		fatal("unable to set exit function");

	if((conf = (struct amselwall_config*)
	       calloc(1, sizeof(struct amselwall_config))) == NULL)
		exit(ENOMEM);

	amwall_conf_init();

	if((ret = amwall_getopt(argc, argv)) < 0) {
		exit(-ret);
	}

	if(conf->flags & AMSELWALL_OPT_QUIET) {
		exit(amwall_parse_config());
	} else {
		if(amwall_parse_config() != 0)
			fatal("failed to parse configuration file");
		
		printf("# "VERSION_NAME "\n", VERSION_CHR);
		printf("# Generated from file: \"%s\"\n", 
		    conf->path == NULL ? AMSELWALL_DEFAULT_CONFIG : conf->path);

		TAILQ_FOREACH(table, &conf->h_head, entries)
			amwall_generate(table);
	}

	exit(ret);
}

/*
 * Check user privileges
 */
int 
amwall_superuser(void)
{
	if(getuid() == 0)
		return(1);

	return(0);
}
