/* -*- knf -*- */
/* $AMSEL: parser.c,v 1.3 2004/02/07 14:25:14 reyk Exp $ */

/*
 * *** AMSEL wall note:
 *     This file is based on OpenBSD bgpd's parse.y
 *     (http://www.openbsd.org/cgi-bin/cvsweb/src/usr.sbin/bgpd/parse.y).
 *     It is licensed under the terms of it's original license:
 *
 * Copyright (c) 2002, 2003, 2004 Henning Brauer <henning@openbsd.org>
 * Copyright (c) 2001 Markus Friedl.  All rights reserved.
 * Copyright (c) 2001 Daniel Hartmeier.  All rights reserved.
 * Copyright (c) 2001 Theo de Raadt.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "main.h"
#include "parser.h"
#include "grammar.h"

struct sym {
	TAILQ_ENTRY(sym) entries;
	int used;
	int persist;
	char *nam;
	char *val;
};
TAILQ_HEAD(symhead, sym) symhead;

int
x_yyerror(const char *fmt, ...)
{
	va_list	ap;
	char *nfmt;

	errors = 1;
	va_start(ap, fmt);
	if(asprintf(&nfmt, "[%s:%d] %s\n", infile, yylval.lineno, fmt) == -1)
		fatal("yyerror: asprintf");
	vprintf(nfmt, ap);
	va_end(ap);
	free(nfmt);
	return (0);
}

int 
amwall_lookup(char *s)
{
	int val = 0;

	if((val = amwall_lookup_keyword(s, keywords, keywords_size)) < 0)
		return(STRING);

	return(val);
}

#define MAXPUSHBACK	128

char	*parsebuf;
int	 parseindex;
char	 pushback_buffer[MAXPUSHBACK];
int	 pushback_index = 0;

int
amwall_lgetc(FILE *f)
{
	int	c, next;

	if (parsebuf) {
		/* Read character from the parsebuffer instead of input. */
		if (parseindex >= 0) {
			c = parsebuf[parseindex++];
			if (c != '\0')
				return (c);
			parsebuf = NULL;
		} else
			parseindex++;
	}

	if (pushback_index)
		return (pushback_buffer[--pushback_index]);

	while ((c = getc(f)) == '\\') {
		next = getc(f);
		if (next != '\n') {
			if (isspace(next))
				yyerror("whitespace after \\");
			ungetc(next, f);
			break;
		}
		yylval.lineno = lineno;
		lineno++;
	}
	if (c == '\t' || c == ' ') {
		/* Compress blanks to a single space. */
		do {
			c = getc(f);
		} while (c == '\t' || c == ' ');
		ungetc(c, f);
		c = ' ';
	}

	return (c);
}

int
amwall_lungetc(int c)
{
	if (c == EOF)
		return (EOF);
	if (parsebuf) {
		parseindex--;
		if (parseindex >= 0)
			return (c);
	}
	if (pushback_index < MAXPUSHBACK-1)
		return (pushback_buffer[pushback_index++] = c);
	else
		return (EOF);
}

int
amwall_findeol(void)
{
	int	c;

	parsebuf = NULL;
	pushback_index = 0;

	/* skip to either EOF or the first real EOL */
	while (1) {
		c = amwall_lgetc(fin);
		if (c == '\n') {
			lineno++;
			break;
		}
		if (c == EOF)
			break;
	}
	return (errors);
}

int
yylex(void)
{
	char	 buf[BUFSIZ];
	char	*p, *val;
	int	 endc, c;
	int	 token;

	memset(buf, 0, sizeof(buf));

top:
	p = buf;
	while ((c = amwall_lgetc(fin)) == ' ')
		; /* nothing */

	yylval.lineno = lineno;
	if (c == '#')
		while ((c = amwall_lgetc(fin)) != '\n' && c != EOF)
			; /* nothing */
	if (c == '$' && parsebuf == NULL) {
		while (1) {
			if ((c = amwall_lgetc(fin)) == EOF)
				return (0);

			if (p + 1 >= buf + sizeof(buf) - 1) {
				yyerror("string too long");
				return (amwall_findeol());
			}
			if (isalnum(c) || c == '_') {
				*p++ = (char)c;
				continue;
			}
			*p = '\0';
			amwall_lungetc(c);
			break;
		}
		val = amwall_symget(buf);
		if (val == NULL) {
			yyerror("macro '%s' not defined", buf);
			return (amwall_findeol());
		}
		parsebuf = val;
		parseindex = 0;
		goto top;
	}

	switch (c) {
	case '\'':
	case '"':
		endc = c;
		while (1) {
			if ((c = amwall_lgetc(fin)) == EOF)
				return (0);
			if (c == endc) {
				*p = '\0';
				break;
			}
			if (c == '\n') {
				lineno++;
				continue;
			}
			if (p + 1 >= buf + sizeof(buf) - 1) {
				yyerror("string too long");
				return (amwall_findeol());
			}
			*p++ = (char)c;
		}
		yylval.v.string = strdup(buf);
		if (yylval.v.string == NULL)
			fatal("yylex: strdup");
		return (STRING);
	}

#define allowed_in_string(x) \
	(isalnum(x) || (ispunct(x) && x != '(' && x != ')' && \
	x != '{' && x != '}' && x != '<' && x != '>' && \
	x != '!' && x != '=' && x != '/' && x != '#' && \
	x != ',' && x != ';' && x != '.'))

	if (isalnum(c) || c == ':' || c == '_') {
		do {
			*p++ = c;
			if ((unsigned)(p-buf) >= sizeof(buf)) {
				yyerror("string too long");
				return (amwall_findeol());
			}
		} while ((c = amwall_lgetc(fin)) != EOF && (allowed_in_string(c)));
		amwall_lungetc(c);
		*p = '\0';

		if((token = amwall_lookup(buf)) == STRING) {
			yylval.v.string = strdup(buf);
			
			if (yylval.v.string == NULL)
				fatal("yylex: strdup");
		}

		return (token);
	}
	if (c == '\n') {
		yylval.lineno = lineno;
		lineno++;
	}
	if (c == EOF)
		return (0);
	return (c);
}

int 
amwall_parse_config(void)
{
	struct sym *sym, *next_sym;

	if(conf->flags & AMSELWALL_OPT_PATH && conf->path != NULL)
		infile = conf->path;
	else
		infile = AMSELWALL_DEFAULT_CONFIG;

	if ((fin = fopen(infile, "r")) == NULL) {
		warn("%s", infile);
		return(-ENOENT);
	}

	TAILQ_INIT(&symhead);

	yyparse();

	fclose(fin);

	/* Free macros and check which have not been used. */
	for (sym = TAILQ_FIRST(&symhead); sym != NULL; sym = next_sym) {
		next_sym = TAILQ_NEXT(sym, entries);
		if (!sym->persist) {
			free(sym->nam);
			free(sym->val);
			TAILQ_REMOVE(&symhead, sym, entries);
			free(sym);
		}
	}

	return(errors ? -1 : 0);
}

int
amwall_symset(const char *nam, const char *val, int persist)
{
	struct sym *sym;

	for (sym = TAILQ_FIRST(&symhead); sym && strcmp(nam, sym->nam);
	     sym = TAILQ_NEXT(sym, entries)); /* nothing */

	if (sym != NULL) {
		if (sym->persist == 1)
			return (0);
		else {
			free(sym->nam);
			free(sym->val);
			TAILQ_REMOVE(&symhead, sym, entries);
			free(sym);
		}
	}
	if ((sym = (struct sym*)calloc(1, sizeof(*sym))) == NULL)
		return (-1);

	sym->nam = strdup(nam);
	if (sym->nam == NULL) {
		free(sym);
		return (-1);
	}
	sym->val = strdup(val);
	if (sym->val == NULL) {
		free(sym->nam);
		free(sym);
		return (-1);
	}
	sym->used = 0;
	sym->persist = persist;
	TAILQ_INSERT_TAIL(&symhead, sym, entries);
	return (0);
}

int
amwall_cmdline_symset(char *s)
{
	char *sym, *val;
	size_t len;
	int ret;

	if ((val = strrchr(s, '=')) == NULL)
		return (-1);

	len = strlen(s) - strlen(val) + 1;
	if ((sym = (char*)calloc(1, len)) == NULL)
		fatal("cmdline_symset: calloc");
	strlcpy(sym, s, len);

	ret = amwall_symset(sym, val + 1, 1);
	free(sym);

	return (ret);
}

char *
amwall_symget(const char *nam)
{
	struct sym	*sym;

	TAILQ_FOREACH(sym, &symhead, entries)
		if (strcmp(nam, sym->nam) == 0) {
			sym->used = 1;
			return (sym->val);
		}
	return (NULL);
}
